/* Copyright 2021 Alibaba Group Holding Limited. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================*/

#ifndef HYBRIDBACKEND_TENSORFLOW_DISTRIBUTE_NCCL_COMM_H_
#define HYBRIDBACKEND_TENSORFLOW_DISTRIBUTE_NCCL_COMM_H_

#if HYBRIDBACKEND_TENSORFLOW

#include <tensorflow/core/framework/resource_mgr.h>
#include <tensorflow/core/framework/tensor.h>
#include <tensorflow/core/lib/core/errors.h>
#include <tensorflow/core/public/version.h>

#if GOOGLE_CUDA
#include <tensorflow/core/common_runtime/gpu/gpu_event_mgr.h>
#include <tensorflow/stream_executor/stream_executor.h>
#endif

#include "hybridbackend/tensorflow/common/stream.h"
#include "hybridbackend/tensorflow/distribute/nccl/types.h"

namespace tensorflow {

#if (TF_MAJOR_VERSION * 1000L + TF_MINOR_VERSION) <= 1008L
namespace se = ::perftools::gputools;
#endif

namespace hybridbackend {

#if HYBRIDBACKEND_NCCL
#if GOOGLE_CUDA

class NcclComm : public ResourceBase {
 public:
  int rank() const { return rank_; }

  int world_size() const { return world_size_; }

  Stream* stream() const { return stream_; }

#if (TF_MAJOR_VERSION * 1000L + TF_MINOR_VERSION) >= 1014L
  string DebugString() const override {
#else
  string DebugString() override {
#endif
    return debug_string_;
  }

  NcclComm();

  ~NcclComm();

  Status Initialize(OpKernelContext* ctx, const string& shared_name, int size,
                    int rank);

  Status Create(const string& id);

  Status Destroy();

  Status CheckAsyncErrors();

  Status UserRank(int* rank);

  // Symmetric collective operations

  Status Allreduce(const Tensor& input, const ncclRedOp_t reduce_op,
                   Tensor* output);

  Status AllreduceN(const std::vector<Tensor>& n_input,
                    const ncclRedOp_t reduce_op,
                    std::vector<Tensor*>* n_output);

  Status Alltoall(const Tensor& input, Tensor* output);

  Status AlltoallN(const std::vector<Tensor>& n_input,
                   std::vector<Tensor*>* n_output);

  Status AlltoallN(const std::vector<Tensor*>& n_input,
                   std::vector<Tensor*>* n_output);

  Status Alltoallv(const Tensor& input, const int32* send_sizes,
                   const int32* recv_sizes, const int64 common_size,
                   Tensor* output);

  Status AlltoallvN(const std::vector<Tensor>& n_input,
                    const std::vector<int32*>& n_send_sizes,
                    const std::vector<int32*>& n_recv_sizes,
                    const std::vector<int64>& n_common_size,
                    std::vector<Tensor*>* n_output);

  Status AlltoallvN(const std::vector<Tensor*>& n_input,
                    const std::vector<int32*>& n_send_sizes,
                    const std::vector<int32*>& n_recv_sizes,
                    const std::vector<int64>& n_common_size,
                    std::vector<Tensor*>* n_output);

  // Asymmetric collective operations

  Status Broadcast(const Tensor& input, const int root_rank, Tensor* output);

  Status Allgather(const Tensor& input, Tensor* output);

  Status Allgatherv(const Tensor& input, const Tensor& host_sizes,
                    Tensor* output);

 private:
  ncclComm_t comm_;
  int world_size_;
  int rank_;
  bool created_;
  string debug_string_;
  Stream* stream_;
  std::mutex mu_;
};

class NcclCommAsyncOp : public AsyncOpKernel {
 public:
  explicit NcclCommAsyncOp(OpKernelConstruction* ctx);

  virtual void ComputeAsyncWithComm(NcclComm* comm, OpKernelContext* ctx,
                                    DoneCallback done) = 0;

  void ComputeAsync(OpKernelContext* ctx,
                    AsyncOpKernel::DoneCallback done) override;
};

#endif
#endif

}  // namespace hybridbackend
}  // namespace tensorflow

#endif

#endif  // HYBRIDBACKEND_TENSORFLOW_DISTRIBUTE_NCCL_COMM_H_
